in(node::Node, graph::DAG) = node in graph.nodes
in(edge::Edge, graph::DAG) = edge in graph.edges

function ==(n1::Node, n2::Node, g::DAG)
   if typeof(n1) != typeof(n2)
      return false
   end
   if !(n1 in g) || !(n2 in g)
      return false
   end

   return n1.task == n2.task && children(n1, g) == children(n2, g)
end

# children = prerequisite nodes, nodes that need to execute before the task, edges point into this task
function children(graph::DAG, node::Node)
   if !(node in graph)
      error("Cannot get children of a node that's not in the given graph")
   end

   result = Vector{Node}()
   for edge in graph.edges
      if (edge.edge[2][] == node)
         push!(result, edge.edge[1][])
      end
   end

   return result
end

# parents = subsequent nodes, nodes that need this node to execute, edges point from this task
function parents(graph::DAG, node::Node)
   if !(node in graph)
      error("Cannot get parents of a node that's not in the given graph")
   end

   result = Vector{Node}()
   for edge in graph.edges
      if (edge.edge[1][] == node)
         push!(result, edge.edge[2][])
      end
   end

   return result
end

is_entry_node(graph::DAG, node::Node) = size(children(graph, node)) == 0
is_exit_node(graph::DAG, node::Node) = size(parents(graph, node)) == 0

function insert_node(graph::DAG, node::Node)
   push!(graph.nodes, node)
   return nothing
end

function insert_edge(graph::DAG, edge::Edge)
   push!(graph.edges, edge)
   return nothing
end

function remove_node(graph::DAG, node::Node)
   deleteat!(graph.nodes, findall(x->x==node, graph.nodes))
   return nothing
end

function remove_edge(graph::DAG, edge::Edge)
   deleteat!(graph.edges, findall(x->x==edge, graph.edges))
   return nothing
end

function data(graph::DAG)
   data_sum::Int64 = 0
   for node in graph.nodes
      data_sum = data_sum + data(node.task)
   end
   return data_sum
end

function compute_effort(graph::DAG)
   compute_effort_sum::Int64 = 0
   for node in graph.nodes
      compute_effort_sum = compute_effort_sum + compute_effort(node.task)
   end
   return compute_effort_sum
end

function compute_intensity(graph::DAG)
   data_sum = data(graph)
   if data_sum == 0
      return typemax(UInt64)
   end
   return compute_effort(graph) / data_sum
end

function show_nodes(io, graph::DAG)
   print(io, "[")
   first = true
   for n in graph.nodes
      if first
         first = false
      else
         print(", ")
      end
      print(io, n)
   end
   print(io, "]")
end

function show_edges(io, graph::DAG)
   print(io, "[")
   first = true
   for e in graph.edges
      if first
         first = false
      else
         print(", ")
      end
      print(io, e)
   end
   print(io, "]")
end

function show(io::IO, graph::DAG)
   println(io, "Graph:")
   print("  Nodes: ")
   show_nodes(io, graph)
   println()
   print("  Edges: ")
   show_edges(io, graph)
   println()
   println("  Total Compute Effort: ", compute_effort(graph))
   println("  Total Data Transfer: ", data(graph))
   println("  Total Compute Intensity: ", compute_intensity(graph))
end
